#ifndef ISERVER_HH
#define ISERVER_HH

#include <sys/epoll.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <iostream>
#include <vector>

#include "imsg.hh"
#include "message.hh"
#include "msgConstants.hh"

class iserver {

	private:
		unsigned int listenPort;
		int listenSocket, epollfdr;
		struct sockaddr_in serverAddress;
		struct sockaddr_in clientAddress;
		struct epoll_event ev;
		struct epoll_event events[MAX_CLIENT_CONNECTIONS];
		std::vector<int> socketfds;
		int inactivity_time_ms;
		int inactivity_time_sec;
		int maxConnections;
		struct timeval tv;
		
	public:
		static const int READ_EVENT		= 1;
		static const int WRITE_EVENT	= 2;
		static const int ERROR_EVENT	= 3;
		static const int STAT_ERR		= 0;
		static const int STAT_RECV		= 1;
	
	iserver(){}
	
	bool createServer(unsigned int p, unsigned int maxConnects) {
		inactivity_time_sec = 0;
		inactivity_time_ms = 0;
		maxConnections = maxConnects;
		listenPort = p;
		listenSocket = socket(AF_INET, SOCK_STREAM, 0);
		if(listenSocket < 0) std::cerr << "error server open socket\n";
		serverAddress.sin_family = AF_INET;
		serverAddress.sin_addr.s_addr = htonl(INADDR_ANY);
		serverAddress.sin_port = htons(listenPort);
		int optval = 1, notcpdelay = 1;
		if(setsockopt(listenSocket, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(int)) == -1) return false;
		if(setsockopt(listenSocket, SOL_TCP, TCP_NODELAY, &notcpdelay, sizeof(int)) == -1) ; //"TCP_NODELAY unavailable";   
		if(::bind(listenSocket, (struct sockaddr *) &serverAddress, sizeof(serverAddress)) < 0) return false;
		if(::listen(listenSocket, maxConnections) < 0) return false;
		epollfdr = epoll_create(maxConnections);
		if(!epollfdr) return false;
		ev.events = EPOLLIN;
		ev.data.fd = listenSocket;
		if(epoll_ctl(epollfdr, EPOLL_CTL_ADD, listenSocket, &ev) < 0) return false;
		return true;
	}
	
	int addClientConnection() {
		int newsockfd;
		socklen_t clientAddressLength = sizeof(clientAddress);
		newsockfd = ::accept(listenSocket, (struct sockaddr *) &clientAddress, &clientAddressLength);
		ev.events = EPOLLIN;
		ev.data.fd = newsockfd;
		if (epoll_ctl(epollfdr, EPOLL_CTL_ADD, newsockfd, &ev) == -1) return -1;
		socketfds.push_back(newsockfd);
		return newsockfd;
	}
	
	void removeClientConnection(int c) {
		std::vector<int>::iterator it;
		for(it = socketfds.begin() ; it != socketfds.end() ; it++) {
			if(*it == c) {
				::close(c);
				socketfds.erase(it);
				epoll_ctl(epollfdr, EPOLL_CTL_DEL, c, NULL);
				break;
			}
		}
	}

	std::vector<std::pair<int, int> > waitReadEvents() {
		std::vector<std::pair<int, int> > resultSet;
		int status = ERROR_EVENT;
		int currentTimeCounter = 0;
		while(true) {
			int nfds = epoll_wait(epollfdr, events, maxConnections, inactivity_time_ms);
			if(nfds == -1) { // error management
				resultSet.push_back(std::make_pair(status, -1));
				return resultSet;
			}
			else if (nfds == 0) { // timeout
				currentTimeCounter++;
				if(currentTimeCounter == inactivity_time_sec) {
					resultSet.push_back(std::make_pair(status, 0));
				    return resultSet; 
				}
			} 
			else {
				for(int i = 0 ; i < nfds ; i++) {
					status = READ_EVENT;
					if((!(events[i].events & EPOLLIN))||(events[i].events & EPOLLHUP)||(events[i].events & EPOLLERR)) 
						status = ERROR_EVENT;
					resultSet.push_back(std::make_pair(status, events[i].data.fd));
				}
				return resultSet;
			}
		}
	}

	bool isMainSocket(int s) {
		return (s == listenSocket);
	}

	void setInactivityTime(int inact) {
		inactivity_time_sec = inact;
		if(inactivity_time_sec == 0) inactivity_time_ms = 0;
		else inactivity_time_ms = 1000;
	}

	bool send(int sock, message data) {
		imsg msg;
		msg.encode(data.data());
		long n, tot = 0;
		long bleft = msg.length();
		while(tot < msg.length()) {
			n = ::send(sock, msg.data() + tot, bleft, MSG_NOSIGNAL);
			if(n < 1) break;
			tot += n;
			bleft -= tot;
		}
		return (tot == msg.length());
	}
		
	std::pair<int, message> receive(int sock) {
			message result;
			imsg msg;
			long res;
			char header[msg.hsize()];
			res = ::recv(sock, header, msg.hsize(), 0);
			if(res != msg.hsize()) return std::make_pair(STAT_ERR, result);
			msg.decodeHeader(header);
			char body[msg.blength()];
			res = ::recv(sock, body, msg.blength(), MSG_WAITALL);
			if(res != msg.blength()) return std::make_pair(STAT_ERR, result);
			msg.decodeBody(body);
			result.decode(msg.body());
			return std::make_pair(STAT_RECV, result);
	}

	void stopServer() {
		if (listenSocket >= 0) {
			for(int i = 0 ; i < socketfds.size() ; i++) ::close(socketfds[i]);
			::close(listenSocket);
			socketfds.clear();
		} else std::cerr << "error at server stop\n";
	}
	
};

#endif

